@@ -37,11 +37,10 @@ module Bunny
   # Channels can be opened either via `Bunny::Session#create_channel` (sufficient in the majority
   # of cases) or by instantiating `Bunny::Channel` directly:
   #
-  # @example Using {Bunny::Session#create_channel}:
-  #   conn = Bunny.new
-  #   conn.start
+  #     conn = Bunny.new
+  #     conn.start
   #
-  #   ch   = conn.create_channel
+  #     ch   = conn.create_channel
   #
   # This will automatically allocate a channel id.
   #
@@ -51,10 +50,8 @@ module Bunny
   # closed, too. Closed channels can no longer be used. Attempts to use them will raise
   # {Bunny::ChannelAlreadyClosed}.
   #
-  # @example
-  #
-  #   ch  = conn.create_channel
-  #   ch.close
+  #     ch = conn.create_channel
+  #     ch.close
   #
   # ## Higher-level API
   #
@@ -157,7 +154,13 @@ module Bunny
     # @return [Hash<String, Bunny::Consumer>] Consumer instances declared on this channel
     attr_reader :consumers
 
+    # @return [Integer] active basic.qos prefetch value
+    attr_reader :prefetch_count
+    # @return [Integer] active basic.qos prefetch global mode
+    attr_reader :prefetch_global
+
     DEFAULT_CONTENT_TYPE = "application/octet-stream".freeze
+    SHORTSTR_LIMIT = 255
 
     # @param [Bunny::Session] connection AMQP 0.9.1 connection
     # @param [Integer] id Channel id, pass nil to make Bunny automatically allocate it
@@ -191,18 +194,19 @@ module Bunny
       @threads_waiting_on_basic_get_continuations = Set.new
 
       @next_publish_seq_no = 0
+      @delivery_tag_offset = 0
 
       @recoveries_counter = Bunny::Concurrent::AtomicFixnum.new(0)
       @uncaught_exception_handler = Proc.new do |e, consumer|
-        @logger.error "Uncaught exception from consumer #{consumer.to_s}: #{e.message}"
+        @logger.error "Uncaught exception from consumer #{consumer.to_s}: #{e.inspect} @ #{e.backtrace[0]}"
       end
     end
 
     attr_reader :recoveries_counter
 
     # @private
-    def read_write_timeout
-      @connection.read_write_timeout
+    def wait_on_continuations_timeout
+      @connection.transport_write_timeout
     end
 
     # Opens the channel and resets its internal state
@@ -227,7 +231,8 @@ module Bunny
     # @api public
     def close
       @connection.close_channel(self)
-      closed!
+      @status = :closed
+      @work_pool.shutdown
       maybe_kill_consumer_work_pool!
     end
 
@@ -243,6 +248,14 @@ module Bunny
       @status == :closed
     end
 
+    def to_s
+      oid = ("0x%x" % (self.object_id << 1))
+      "<#{self.class.name}:#{oid} number=#{@channel.id} @open=#{open?} connection=#{@connection.to_s}>"
+    end
+
+    def inspect
+      to_s
+    end
 
     #
     # @group Backwards compatibility with 0.8.0
@@ -353,7 +366,7 @@ module Bunny
     # @see http://rubybunny.info/articles/exchanges.html Exchanges and Publishing guide
     # @api public
     def default_exchange
-      self.direct(AMQ::Protocol::EMPTY_STRING, :no_declare => true)
+      Exchange.default(self)
     end
 
     # Declares a headers exchange or looks it up in the cache of previously
@@ -399,22 +412,21 @@ module Bunny
       register_queue(q)
     end
 
+    # Declares a new server-named queue that is automatically deleted when the
+    # connection is closed.
+    #
+    # @return [Bunny::Queue] Queue that was declared
+    # @see #queue
+    # @api public
+    def temporary_queue(opts = {})
+      queue("", opts.merge(:exclusive => true))
+    end
+
     # @endgroup
 
 
     # @group QoS and Flow Control
 
-    # Sets how many messages will be given to consumers on this channel before they
-    # have to acknowledge or reject one of the previously consumed messages
-    #
-    # @param [Integer] prefetch_count Prefetch (QoS setting) for this channel
-    # @see http://rubybunny.info/articles/exchanges.html Exchanges and Publishing guide
-    # @see http://rubybunny.info/articles/queues.html Queues and Consumers guide
-    # @api public
-    def prefetch(prefetch_count)
-      self.basic_qos(prefetch_count, false)
-    end
-
     # Flow control. When set to false, RabbitMQ will stop delivering messages on this
     # channel.
     #
@@ -447,9 +459,7 @@ module Bunny
     # @see http://rubybunny.info/articles/queues.html Queues and Consumers guide
     # @api public
     def reject(delivery_tag, requeue = false)
-      guarding_against_stale_delivery_tags(delivery_tag) do
-        basic_reject(delivery_tag.to_i, requeue)
-      end
+      basic_reject(delivery_tag.to_i, requeue)
     end
 
     # Acknowledges a message. Acknowledged messages are completely removed from the queue.
@@ -460,9 +470,7 @@ module Bunny
     # @see http://rubybunny.info/articles/queues.html Queues and Consumers guide
     # @api public
     def ack(delivery_tag, multiple = false)
-      guarding_against_stale_delivery_tags(delivery_tag) do
-        basic_ack(delivery_tag.to_i, multiple)
-      end
+      basic_ack(delivery_tag.to_i, multiple)
     end
     alias acknowledge ack
 
@@ -477,9 +485,7 @@ module Bunny
     # @see http://rubybunny.info/articles/queues.html Queues and Consumers guide
     # @api public
     def nack(delivery_tag, multiple = false, requeue = false)
-      guarding_against_stale_delivery_tags(delivery_tag) do
-        basic_nack(delivery_tag.to_i, multiple, requeue)
-      end
+      basic_nack(delivery_tag.to_i, multiple, requeue)
     end
 
     # @endgroup
@@ -516,6 +522,7 @@ module Bunny
     # @api public
     def basic_publish(payload, exchange, routing_key, opts = {})
       raise_if_no_longer_open!
+      raise ArgumentError, "routing key cannot be longer than #{SHORTSTR_LIMIT} characters" if routing_key && routing_key.size > SHORTSTR_LIMIT
 
       exchange_name = if exchange.respond_to?(:name)
                         exchange.name
@@ -534,8 +541,10 @@ module Bunny
       opts[:priority]      ||= 0
 
       if @next_publish_seq_no > 0
-        @unconfirmed_set.add(@next_publish_seq_no)
-        @next_publish_seq_no += 1
+        @unconfirmed_set_mutex.synchronize do
+          @unconfirmed_set.add(@next_publish_seq_no)
+          @next_publish_seq_no += 1
+        end
       end
 
       frames = AMQ::Protocol::Basic::Publish.encode(@id,
@@ -546,7 +555,7 @@ module Bunny
         opts[:mandatory],
         false,
         @connection.frame_max)
-      @connection.send_frameset_without_timeout(frames, self)
+      @connection.send_frameset(frames, self)
 
       self
     end
@@ -558,7 +567,8 @@ module Bunny
     # @param [String] queue Queue name
     # @param [Hash] opts Options
     #
-    # @option opts [Boolean] :ack (true) Will this message be acknowledged manually?
+    # @option opts [Boolean] :ack (true) [DEPRECATED] Use :manual_ack instead
+    # @option opts [Boolean] :manual_ack (true) Will this message be acknowledged manually?
     #
     # @return [Array] A triple of delivery info, message properties and message content
     #
@@ -567,15 +577,20 @@ module Bunny
     #   conn.start
     #   ch   = conn.create_channel
     #   # here we assume the queue already exists and has messages
-    #   delivery_info, properties, payload = ch.basic_get("bunny.examples.queue1", :ack => true)
+    #   delivery_info, properties, payload = ch.basic_get("bunny.examples.queue1", :manual_ack => true)
     #   ch.acknowledge(delivery_info.delivery_tag)
     # @see Bunny::Queue#pop
     # @see http://rubybunny.info/articles/queues.html Queues and Consumers guide
     # @api public
-    def basic_get(queue, opts = {:ack => true})
+    def basic_get(queue, opts = {:manual_ack => true})
       raise_if_no_longer_open!
 
-      @connection.send_frame(AMQ::Protocol::Basic::Get.encode(@id, queue, !(opts[:ack])))
+      unless opts[:ack].nil?
+        warn "[DEPRECATION] `:ack` is deprecated.  Please use `:manual_ack` instead."
+        opts[:manual_ack] = opts[:ack]
+      end
+
+      @connection.send_frame(AMQ::Protocol::Basic::Get.encode(@id, queue, !(opts[:manual_ack])))
       # this is a workaround for the edge case when basic_get is called in a tight loop
       # and network goes down we need to perform recovery. The problem is, basic_get will
       # keep blocking the thread that calls it without clear way to constantly unblock it
@@ -583,40 +598,54 @@ module Bunny
       # implementation (and even more correct and convenient ones, such as wait/notify, should
       # we implement them). So we return a triple of nils immediately which apps should be
       # able to handle anyway as "got no message, no need to act". MK.
-      @last_basic_get_response = if @connection.open?
-                                   wait_on_basic_get_continuations
-                                 else
-                                   [nil, nil, nil]
-                                 end
+      last_basic_get_response = if @connection.open?
+                                  wait_on_basic_get_continuations
+                                else
+                                  [nil, nil, nil]
+                                end
 
       raise_if_continuation_resulted_in_a_channel_error!
-      @last_basic_get_response
+      last_basic_get_response
     end
 
+    # prefetch_count is of type short in the protocol. MK.
+    MAX_PREFETCH_COUNT = (2 ** 16) - 1
+
     # Controls message delivery rate using basic.qos AMQP 0.9.1 method.
     #
     # @param [Integer] prefetch_count How many messages can consumers on this channel be given at a time
     #                                 (before they have to acknowledge or reject one of the earlier received messages)
-    # @param [Boolean] global (false) Ignored, as it is not supported by RabbitMQ
+    # @param [Boolean] global
+    #   Whether to use global mode for prefetch:
+    #   - +false+: per-consumer
+    #   - +true+:  per-channel
+    #   Note that the default value (+false+) hasn't actually changed, but
+    #   previous documentation described that as meaning per-channel and
+    #   unsupported in RabbitMQ, whereas it now actually appears to mean
+    #   per-consumer and supported
+    #   (https://www.rabbitmq.com/consumer-prefetch.html).
     # @return [AMQ::Protocol::Basic::QosOk] RabbitMQ response
     # @see Bunny::Channel#prefetch
     # @see http://rubybunny.info/articles/queues.html Queues and Consumers guide
     # @api public
-    def basic_qos(prefetch_count, global = false)
-      raise ArgumentError.new("prefetch count must be a positive integer, given: #{prefetch_count}") if prefetch_count < 0
+    def basic_qos(count, global = false)
+      raise ArgumentError.new("prefetch count must be a positive integer, given: #{count}") if count < 0
+      raise ArgumentError.new("prefetch count must be no greater than #{MAX_PREFETCH_COUNT}, given: #{count}") if count > MAX_PREFETCH_COUNT
       raise_if_no_longer_open!
 
-      @connection.send_frame(AMQ::Protocol::Basic::Qos.encode(@id, 0, prefetch_count, global))
+      @connection.send_frame(AMQ::Protocol::Basic::Qos.encode(@id, 0, count, global))
 
-      Bunny::Timeout.timeout(read_write_timeout, ClientTimeout) do
+      Bunny::Timeout.timeout(wait_on_continuations_timeout, ClientTimeout) do
         @last_basic_qos_ok = wait_on_continuations
       end
       raise_if_continuation_resulted_in_a_channel_error!
 
-      @prefetch_count = prefetch_count
+      @prefetch_count  = count
+      @prefetch_global = global
 
       @last_basic_qos_ok
     end
+    alias prefetch basic_qos
 
     # Redeliver unacknowledged messages
     #
@@ -627,7 +656,7 @@ module Bunny
       raise_if_no_longer_open!
 
       @connection.send_frame(AMQ::Protocol::Basic::Recover.encode(@id, requeue))
-      Bunny::Timeout.timeout(read_write_timeout, ClientTimeout) do
+      Bunny::Timeout.timeout(wait_on_continuations_timeout, ClientTimeout) do
         @last_basic_recover_ok = wait_on_continuations
       end
       raise_if_continuation_resulted_in_a_channel_error!
@@ -657,7 +686,7 @@ module Bunny
     #
     #   ch    = conn.create_channel
     #   q.subscribe do |delivery_info, properties, payload|
-    #     # requeue the message
+    #     # reject the message
     #     ch.basic_reject(delivery_info.delivery_tag, false)
     #   end
     #
@@ -667,17 +696,19 @@ module Bunny
     #
     #   ch    = conn.create_channel
     #   # we assume the queue exists and has messages
-    #   delivery_info, properties, payload = ch.basic_get("bunny.examples.queue3", :ack => true)
+    #   delivery_info, properties, payload = ch.basic_get("bunny.examples.queue3", :manual_ack => true)
     #   ch.basic_reject(delivery_info.delivery_tag, true)
     #
     # @see Bunny::Channel#basic_nack
     # @see http://rubybunny.info/articles/queues.html Queues and Consumers guide
     # @api public
-    def basic_reject(delivery_tag, requeue)
-      raise_if_no_longer_open!
-      @connection.send_frame(AMQ::Protocol::Basic::Reject.encode(@id, delivery_tag, requeue))
+    def basic_reject(delivery_tag, requeue = false)
+      guarding_against_stale_delivery_tags(delivery_tag) do
+        raise_if_no_longer_open!
+        @connection.send_frame(AMQ::Protocol::Basic::Reject.encode(@id, delivery_tag, requeue))
 
-      nil
+        nil
+      end
     end
 
     # Acknowledges a delivery (message).
@@ -693,7 +724,7 @@ module Bunny
     #   ch    = conn.create_channel
     #   q.subscribe do |delivery_info, properties, payload|
     #     # requeue the message
-    #     ch.basic_ack(delivery_info.delivery_tag)
+    #     ch.basic_ack(delivery_info.delivery_tag.to_i)
     #   end
     #
     # @example Ack a message fetched via basic.get
@@ -702,8 +733,8 @@ module Bunny
     #
     #   ch    = conn.create_channel
     #   # we assume the queue exists and has messages
-    #   delivery_info, properties, payload = ch.basic_get("bunny.examples.queue3", :ack => true)
-    #   ch.basic_ack(delivery_info.delivery_tag)
+    #   delivery_info, properties, payload = ch.basic_get("bunny.examples.queue3", :manual_ack => true)
+    #   ch.basic_ack(delivery_info.delivery_tag.to_i)
     #
     # @example Ack multiple messages fetched via basic.get
     #   conn  = Bunny.new
@@ -711,20 +742,21 @@ module Bunny
     #
     #   ch    = conn.create_channel
     #   # we assume the queue exists and has messages
-    #   _, _, payload1 = ch.basic_get("bunny.examples.queue3", :ack => true)
-    #   _, _, payload2 = ch.basic_get("bunny.examples.queue3", :ack => true)
-    #   delivery_info, properties, payload3 = ch.basic_get("bunny.examples.queue3", :ack => true)
+    #   _, _, payload1 = ch.basic_get("bunny.examples.queue3", :manual_ack => true)
+    #   _, _, payload2 = ch.basic_get("bunny.examples.queue3", :manual_ack => true)
+    #   delivery_info, properties, payload3 = ch.basic_get("bunny.examples.queue3", :manual_ack => true)
     #   # ack all fetched messages up to payload3
-    #   ch.basic_ack(delivery_info.delivery_tag, true)
+    #   ch.basic_ack(delivery_info.delivery_tag.to_i, true)
     #
     # @see http://rubybunny.info/articles/queues.html Queues and Consumers guide
-    # @see #basic_ack_known_delivery_tag
     # @api public
-    def basic_ack(delivery_tag, multiple)
-      raise_if_no_longer_open!
-      @connection.send_frame(AMQ::Protocol::Basic::Ack.encode(@id, delivery_tag, multiple))
+    def basic_ack(delivery_tag, multiple = false)
+      guarding_against_stale_delivery_tags(delivery_tag) do
+        raise_if_no_longer_open!
+        @connection.send_frame(AMQ::Protocol::Basic::Ack.encode(@id, delivery_tag, multiple))
 
-      nil
+        nil
+      end
     end
 
     # Rejects or requeues messages just like {Bunny::Channel#basic_reject} but can do so
@@ -761,7 +793,7 @@ module Bunny
     #
     #   ch    = conn.create_channel
     #   # we assume the queue exists and has messages
-    #   delivery_info, properties, payload = ch.basic_get("bunny.examples.queue3", :ack => true)
+    #   delivery_info, properties, payload = ch.basic_get("bunny.examples.queue3", :manual_ack => true)
     #   ch.basic_nack(delivery_info.delivery_tag, false, true)
     #
     #
@@ -771,9 +803,9 @@ module Bunny
     #
     #   ch    = conn.create_channel
     #   # we assume the queue exists and has messages
-    #   _, _, payload1 = ch.basic_get("bunny.examples.queue3", :ack => true)
-    #   _, _, payload2 = ch.basic_get("bunny.examples.queue3", :ack => true)
-    #   delivery_info, properties, payload3 = ch.basic_get("bunny.examples.queue3", :ack => true)
+    #   _, _, payload1 = ch.basic_get("bunny.examples.queue3", :manual_ack => true)
+    #   _, _, payload2 = ch.basic_get("bunny.examples.queue3", :manual_ack => true)
+    #   delivery_info, properties, payload3 = ch.basic_get("bunny.examples.queue3", :manual_ack => true)
     #   # requeue all fetched messages up to payload3
     #   ch.basic_nack(delivery_info.delivery_tag, true, true)
     #
@@ -781,13 +813,15 @@ module Bunny
     # @see http://rubybunny.info/articles/extensions.html RabbitMQ Extensions guide
     # @api public
     def basic_nack(delivery_tag, multiple = false, requeue = false)
-      raise_if_no_longer_open!
-      @connection.send_frame(AMQ::Protocol::Basic::Nack.encode(@id,
-          delivery_tag,
-          multiple,
-          requeue))
+      guarding_against_stale_delivery_tags(delivery_tag) do
+        raise_if_no_longer_open!
+        @connection.send_frame(AMQ::Protocol::Basic::Nack.encode(@id,
+                                                                 delivery_tag,
+                                                                 multiple,
+                                                                 requeue))
 
-      nil
+        nil
+      end
     end
 
     # Registers a consumer for queue. Delivered messages will be handled with the block
@@ -829,7 +863,7 @@ module Bunny
           arguments))
 
       begin
-        Bunny::Timeout.timeout(read_write_timeout, ClientTimeout) do
+        Bunny::Timeout.timeout(wait_on_continuations_timeout, ClientTimeout) do
           @last_basic_consume_ok = wait_on_continuations
         end
       rescue Exception => e
@@ -879,7 +913,7 @@ module Bunny
           consumer.arguments))
 
       begin
-        Bunny::Timeout.timeout(read_write_timeout, ClientTimeout) do
+        Bunny::Timeout.timeout(wait_on_continuations_timeout, ClientTimeout) do
           @last_basic_consume_ok = wait_on_continuations
         end
       rescue Exception => e
@@ -914,11 +948,13 @@ module Bunny
     def basic_cancel(consumer_tag)
       @connection.send_frame(AMQ::Protocol::Basic::Cancel.encode(@id, consumer_tag, false))
 
-      Bunny::Timeout.timeout(read_write_timeout, ClientTimeout) do
+      Bunny::Timeout.timeout(wait_on_continuations_timeout, ClientTimeout) do
         @last_basic_cancel_ok = wait_on_continuations
       end
 
-      maybe_kill_consumer_work_pool! unless any_consumers?
+      # reduces thread usage for channels that don't have any
+      # consumers
+      @work_pool.shutdown(true) unless self.any_consumers?
 
       @last_basic_cancel_ok
     end
@@ -988,7 +1024,7 @@ module Bunny
           opts[:if_unused],
           opts[:if_empty],
           false))
-      Bunny::Timeout.timeout(read_write_timeout, ClientTimeout) do
+      Bunny::Timeout.timeout(wait_on_continuations_timeout, ClientTimeout) do
         @last_queue_delete_ok = wait_on_continuations
       end
       raise_if_continuation_resulted_in_a_channel_error!
@@ -1008,7 +1044,7 @@ module Bunny
 
       @connection.send_frame(AMQ::Protocol::Queue::Purge.encode(@id, name, false))
 
-      Bunny::Timeout.timeout(read_write_timeout, ClientTimeout) do
+      Bunny::Timeout.timeout(wait_on_continuations_timeout, ClientTimeout) do
         @last_queue_purge_ok = wait_on_continuations
       end
       raise_if_continuation_resulted_in_a_channel_error!
@@ -1041,10 +1077,10 @@ module Bunny
       @connection.send_frame(AMQ::Protocol::Queue::Bind.encode(@id,
           name,
           exchange_name,
-          opts[:routing_key],
+          (opts[:routing_key] || opts[:key]),
           false,
           opts[:arguments]))
-      Bunny::Timeout.timeout(read_write_timeout, ClientTimeout) do
+      Bunny::Timeout.timeout(wait_on_continuations_timeout, ClientTimeout) do
         @last_queue_bind_ok = wait_on_continuations
       end
 
@@ -1079,7 +1115,7 @@ module Bunny
           exchange_name,
           opts[:routing_key],
           opts[:arguments]))
-      Bunny::Timeout.timeout(read_write_timeout, ClientTimeout) do
+      Bunny::Timeout.timeout(wait_on_continuations_timeout, ClientTimeout) do
         @last_queue_unbind_ok = wait_on_continuations
       end
 
@@ -1095,6 +1131,7 @@ module Bunny
     # Declares a echange using echange.declare AMQP 0.9.1 method.
     #
     # @param [String] name Exchange name
+    # @param [String,Symbol] type Exchange type, e.g. :fanout or :topic
     # @param [Hash] opts Exchange properties
     #
     # @option opts [Boolean] durable (false)     Should information about this echange be persisted to disk so that it
@@ -1118,7 +1155,7 @@ module Bunny
           opts.fetch(:internal, false),
           false, # nowait
           opts[:arguments]))
-      Bunny::Timeout.timeout(read_write_timeout, ClientTimeout) do
+      Bunny::Timeout.timeout(wait_on_continuations_timeout, ClientTimeout) do
         @last_exchange_declare_ok = wait_on_continuations
       end
 
@@ -1143,7 +1180,7 @@ module Bunny
           name,
           opts[:if_unused],
           false))
-      Bunny::Timeout.timeout(read_write_timeout, ClientTimeout) do
+      Bunny::Timeout.timeout(wait_on_continuations_timeout, ClientTimeout) do
         @last_exchange_delete_ok = wait_on_continuations
       end
 
@@ -1187,7 +1224,7 @@ module Bunny
           opts[:routing_key],
           false,
           opts[:arguments]))
-      Bunny::Timeout.timeout(read_write_timeout, ClientTimeout) do
+      Bunny::Timeout.timeout(wait_on_continuations_timeout, ClientTimeout) do
         @last_exchange_bind_ok = wait_on_continuations
       end
 
@@ -1231,7 +1268,7 @@ module Bunny
           opts[:routing_key],
           false,
           opts[:arguments]))
-      Bunny::Timeout.timeout(read_write_timeout, ClientTimeout) do
+      Bunny::Timeout.timeout(wait_on_continuations_timeout, ClientTimeout) do
         @last_exchange_unbind_ok = wait_on_continuations
       end
 
@@ -1259,7 +1296,7 @@ module Bunny
       raise_if_no_longer_open!
 
       @connection.send_frame(AMQ::Protocol::Channel::Flow.encode(@id, active))
-      Bunny::Timeout.timeout(read_write_timeout, ClientTimeout) do
+      Bunny::Timeout.timeout(wait_on_continuations_timeout, ClientTimeout) do
         @last_channel_flow_ok = wait_on_continuations
       end
       raise_if_continuation_resulted_in_a_channel_error!
@@ -1280,7 +1317,7 @@ module Bunny
       raise_if_no_longer_open!
 
       @connection.send_frame(AMQ::Protocol::Tx::Select.encode(@id))
-      Bunny::Timeout.timeout(read_write_timeout, ClientTimeout) do
+      Bunny::Timeout.timeout(wait_on_continuations_timeout, ClientTimeout) do
         @last_tx_select_ok = wait_on_continuations
       end
       raise_if_continuation_resulted_in_a_channel_error!
@@ -1296,7 +1333,7 @@ module Bunny
       raise_if_no_longer_open!
 
       @connection.send_frame(AMQ::Protocol::Tx::Commit.encode(@id))
-      Bunny::Timeout.timeout(read_write_timeout, ClientTimeout) do
+      Bunny::Timeout.timeout(wait_on_continuations_timeout, ClientTimeout) do
         @last_tx_commit_ok = wait_on_continuations
       end
       raise_if_continuation_resulted_in_a_channel_error!
@@ -1311,7 +1348,7 @@ module Bunny
       raise_if_no_longer_open!
 
       @connection.send_frame(AMQ::Protocol::Tx::Rollback.encode(@id))
-      Bunny::Timeout.timeout(read_write_timeout, ClientTimeout) do
+      Bunny::Timeout.timeout(wait_on_continuations_timeout, ClientTimeout) do
         @last_tx_rollback_ok = wait_on_continuations
       end
       raise_if_continuation_resulted_in_a_channel_error!
@@ -1319,6 +1356,11 @@ module Bunny
       @last_tx_rollback_ok
     end
 
+    # @return [Boolean] true if this channel has transactions enabled
+    def using_tx?
+      !!@tx_mode
+    end
+
     # @endgroup
 
 
@@ -1330,6 +1372,7 @@ module Bunny
     def using_publisher_confirmations?
       @next_publish_seq_no > 0
     end
+    alias using_publisher_confirms? using_publisher_confirmations?
 
     # Enables publisher confirms for the channel.
     # @return [AMQ::Protocol::Confirm::SelectOk] RabbitMQ response
@@ -1346,33 +1389,32 @@ module Bunny
         @unconfirmed_set        = Set.new
         @nacked_set             = Set.new
         @next_publish_seq_no    = 1
+        @only_acks_received = true
       end
 
       @confirms_callback = callback
 
       @connection.send_frame(AMQ::Protocol::Confirm::Select.encode(@id, false))
-      Bunny::Timeout.timeout(read_write_timeout, ClientTimeout) do
+      Bunny::Timeout.timeout(wait_on_continuations_timeout, ClientTimeout) do
         @last_confirm_select_ok = wait_on_continuations
       end
-      @confirm_mode = true
       raise_if_continuation_resulted_in_a_channel_error!
       @last_confirm_select_ok
     end
 
     # Blocks calling thread until confirms are received for all
-    # currently unacknowledged published messages.
+    # currently unacknowledged published messages. Returns immediately
+    # if there are no outstanding confirms.
     #
-    # @return [Boolean] true if all messages were acknowledged positively, false otherwise
+    # @return [Boolean] true if all messages were acknowledged positively since the last time this method was called, false otherwise
     # @see #confirm_select
     # @see #unconfirmed_set
     # @see #nacked_set
     # @see http://rubybunny.info/articles/extensions.html RabbitMQ Extensions guide
     # @api public
     def wait_for_confirms
-      @only_acks_received = true
       wait_on_confirms_continuations
-
-      @only_acks_received
+      read_and_reset_only_acks_received
     end
 
     # @endgroup
@@ -1428,7 +1470,7 @@ module Bunny
     #
     # @api plugin
     def recover_from_network_failure
-      @logger.debug "Recovering channel #{@id} after network failure"
+      @logger.debug { "Recovering channel #{@id} after network failure" }
       release_all_continuations
 
       recover_prefetch_setting
@@ -1446,7 +1488,7 @@ module Bunny
     #
     # @api plugin
     def recover_prefetch_setting
-      basic_qos(@prefetch_count) if @prefetch_count
+      basic_qos(@prefetch_count, @prefetch_global) if @prefetch_count
     end
 
     # Recovers publisher confirms mode. Used by the Automatic Network Failure
@@ -1454,7 +1496,11 @@ module Bunny
     #
     # @api plugin
     def recover_confirm_mode
-      confirm_select if @confirm_mode
+      if using_publisher_confirmations?
+        @unconfirmed_set.clear
+        @delivery_tag_offset = @next_publish_seq_no - 1
+        confirm_select(@confirms_callback)
+      end
     end
 
     # Recovers transaction mode. Used by the Automatic Network Failure
@@ -1481,7 +1527,7 @@ module Bunny
     # @api plugin
     def recover_queues
       @queues.values.dup.each do |q|
-        @logger.debug "Recovering queue #{q.name}"
+        @logger.debug { "Recovering queue #{q.name}" }
         q.recover_from_network_failure
       end
     end
@@ -1492,7 +1538,7 @@ module Bunny
     # @api plugin
     def recover_consumers
       unless @consumers.empty?
-        @work_pool = ConsumerWorkPool.new(@work_pool.size)
+        @work_pool = ConsumerWorkPool.new(@work_pool.size, @work_pool.abort_on_exception)
         @work_pool.start
       end
       @consumers.values.dup.each do |c|
@@ -1505,6 +1551,16 @@ module Bunny
       @recoveries_counter.increment
     end
 
+    # @api public
+    def recover_cancelled_consumers!
+      @recover_cancelled_consumers = true
+    end
+
+    # @api public
+    def recovers_cancelled_consumers?
+      !!@recover_cancelled_consumers
+    end
+
     # @endgroup
 
 
@@ -1543,7 +1599,7 @@ module Bunny
 
     # @private
     def handle_method(method)
-      @logger.debug "Channel#handle_frame on channel #{@id}: #{method.inspect}"
+      @logger.debug { "Channel#handle_frame on channel #{@id}: #{method.inspect}" }
       case method
       when AMQ::Protocol::Queue::DeclareOk then
         @continuations.push(method)
@@ -1575,8 +1631,15 @@ module Bunny
         if consumer = @consumers[method.consumer_tag]
           @work_pool.submit do
             begin
-              @consumers.delete(method.consumer_tag)
-              consumer.handle_cancellation(method)
+              if recovers_cancelled_consumers?
+                consumer.handle_cancellation(method)
+                @logger.info "Automatically recovering cancelled consumer #{consumer.consumer_tag} on queue #{consumer.queue_name}"
+
+                consume_with(consumer)
+              else
+                @consumers.delete(method.consumer_tag)
+                consumer.handle_cancellation(method)
+              end
             rescue Exception => e
               @logger.error "Got exception when notifying consumer #{method.consumer_tag} about cancellation!"
               @uncaught_exception_handler.call(e, consumer) if @uncaught_exception_handler
@@ -1661,28 +1724,26 @@ module Bunny
     end
 
     # @private
-    def handle_ack_or_nack(delivery_tag, multiple, nack)
-      if nack
-        cloned_set = @unconfirmed_set.clone
-        if multiple
-          cloned_set.keep_if { |i| i <= delivery_tag }
-          @nacked_set.merge(cloned_set)
-        else
-          @nacked_set.add(delivery_tag)
-        end
-      end
-
-      if multiple
-        @unconfirmed_set.delete_if { |i| i <= delivery_tag }
-      else
-        @unconfirmed_set.delete(delivery_tag)
-      end
+    def handle_ack_or_nack(delivery_tag_before_offset, multiple, nack)
+      delivery_tag          = delivery_tag_before_offset + @delivery_tag_offset
+      confirmed_range_start = multiple ? @delivery_tag_offset + 1 : delivery_tag
+      confirmed_range_end   = delivery_tag
+      confirmed_range       = (confirmed_range_start..confirmed_range_end)
 
       @unconfirmed_set_mutex.synchronize do
+        if nack
+          @nacked_set.merge(@unconfirmed_set & confirmed_range)
+        end
+
+        @unconfirmed_set.subtract(confirmed_range)
+
         @only_acks_received = (@only_acks_received && !nack)
 
         @confirms_continuations.push(true) if @unconfirmed_set.empty?
-        @confirms_callback.call(delivery_tag, multiple, nack) if @confirms_callback
+
+        if @confirms_callback
+          confirmed_range.each { |tag| @confirms_callback.call(tag, false, nack) }
+        end
       end
     end
 
@@ -1724,22 +1785,37 @@ module Bunny
 
     # @private
     def wait_on_confirms_continuations
+      raise_if_no_longer_open!
+
       if @connection.threaded
         t = Thread.current
         @threads_waiting_on_confirms_continuations << t
 
         begin
-          @confirms_continuations.poll(@connection.continuation_timeout)
+          while @unconfirmed_set_mutex.synchronize { !@unconfirmed_set.empty? }
+            @confirms_continuations.poll(@connection.continuation_timeout)
+          end
         ensure
           @threads_waiting_on_confirms_continuations.delete(t)
         end
       else
-        connection.reader_loop.run_once until @confirms_continuations.length > 0
-
-        @confirms_continuations.pop
+        unless @unconfirmed_set.empty?
+          connection.reader_loop.run_once until @confirms_continuations.length > 0
+          @confirms_continuations.pop
+        end
       end
     end
 
+    # @private
+    def read_and_reset_only_acks_received
+      @unconfirmed_set_mutex.synchronize do
+        result = @only_acks_received
+        @only_acks_received = true
+        result
+      end
+    end
+
+
     # Releases all continuations. Used by automatic network recovery.
     # @private
     def release_all_continuations
