@@ -46,11 +46,7 @@ module Bunny
     CONNECT_TIMEOUT   = Transport::DEFAULT_CONNECTION_TIMEOUT
 
     # @private
-    DEFAULT_CONTINUATION_TIMEOUT = if RUBY_VERSION.to_f < 1.9
-                                     8000
-                                   else
-                                     4000
-                                   end
+    DEFAULT_CONTINUATION_TIMEOUT = 15000
 
     # RabbitMQ client metadata
     DEFAULT_CLIENT_PROPERTIES = {
@@ -82,23 +78,25 @@ module Bunny
 
     # @return [Bunny::Transport]
     attr_reader :transport
-    attr_reader :status, :host, :port, :heartbeat, :user, :pass, :vhost, :frame_max, :channel_max, :threaded
+    attr_reader :status, :port, :heartbeat, :user, :pass, :vhost, :frame_max, :channel_max, :threaded
     attr_reader :server_capabilities, :server_properties, :server_authentication_mechanisms, :server_locales
-    attr_reader :default_channel
     attr_reader :channel_id_allocator
     # Authentication mechanism, e.g. "PLAIN" or "EXTERNAL"
     # @return [String]
     attr_reader :mechanism
     # @return [Logger]
     attr_reader :logger
-    # @return [Integer] Timeout for blocking protocol operations (queue.declare, queue.bind, etc), in milliseconds. Default is 4000.
+    # @return [Integer] Timeout for blocking protocol operations (queue.declare, queue.bind, etc), in milliseconds. Default is 15000.
     attr_reader :continuation_timeout
+    attr_reader :network_recovery_interval
 
 
     # @param [String, Hash] connection_string_or_opts Connection string or a hash of connection options
     # @param [Hash] optz Extra options not related to connection
     #
     # @option connection_string_or_opts [String] :host ("127.0.0.1") Hostname or IP address to connect to
+    # @option connection_string_or_opts [Array<String>] :hosts (["127.0.0.1"]) list of hostname or IP addresses to select hostname from when connecting
+    # @option connection_string_or_opts [Array<String>] :addresses (["127.0.0.1:5672"]) list of addresses to select hostname and port from when connecting
     # @option connection_string_or_opts [Integer] :port (5672) Port RabbitMQ listens on
     # @option connection_string_or_opts [String] :username ("guest") Username
     # @option connection_string_or_opts [String] :password ("guest") Password
@@ -109,8 +107,18 @@ module Bunny
     # @option connection_string_or_opts [String] :tls_cert (nil) Path to client TLS/SSL certificate file (.pem)
     # @option connection_string_or_opts [String] :tls_key (nil) Path to client TLS/SSL private key file (.pem)
     # @option connection_string_or_opts [Array<String>] :tls_ca_certificates Array of paths to TLS/SSL CA files (.pem), by default detected from OpenSSL configuration
-    # @option connection_string_or_opts [Integer] :continuation_timeout (4000) Timeout for client operations that expect a response (e.g. {Bunny::Queue#get}), in milliseconds.
+    # @option connection_string_or_opts [String] :verify_peer (true) Whether TLS peer verification should be performed
+    # @option connection_string_or_opts [Keyword] :tls_version (negotiated) What TLS version should be used (:TLSv1, :TLSv1_1, or :TLSv1_2)
+    # @option connection_string_or_opts [Integer] :continuation_timeout (15000) Timeout for client operations that expect a response (e.g. {Bunny::Queue#get}), in milliseconds.
     # @option connection_string_or_opts [Integer] :connection_timeout (5) Timeout in seconds for connecting to the server.
+    # @option connection_string_or_opts [Proc] :hosts_shuffle_strategy A Proc that reorders a list of host strings, defaults to Array#shuffle
+    # @option connection_string_or_opts [Logger] :logger The logger.  If missing, one is created using :log_file and :log_level.
+    # @option connection_string_or_opts [IO, String] :log_file The file or path to use when creating a logger.  Defaults to STDOUT.
+    # @option connection_string_or_opts [IO, String] :logfile DEPRECATED: use :log_file instead.  The file or path to use when creating a logger.  Defaults to STDOUT.
+    # @option connection_string_or_opts [Integer] :log_level The log level to use when creating a logger.  Defaults to LOGGER::WARN
+    # @option connection_string_or_opts [Boolean] :automatically_recover (true) Should automatically recover from network failures?
+    # @option connection_string_or_opts [Integer] :recovery_attempts (nil) Max number of recovery attempts, nil means forever, 0 means never
+    # @option connection_string_or_opts [Boolean] :recover_from_connection_close (true) Recover from server-sent connection.close
     #
     # @option optz [String] :auth_mechanism ("PLAIN") Authentication mechanism, PLAIN or EXTERNAL
     # @option optz [String] :locale ("PLAIN") Locale RabbitMQ should use
@@ -118,26 +126,37 @@ module Bunny
     # @see http://rubybunny.info/articles/connecting.html Connecting to RabbitMQ guide
     # @see http://rubybunny.info/articles/tls.html TLS/SSL guide
     # @api public
-    def initialize(connection_string_or_opts = Hash.new, optz = Hash.new)
-      opts = case (ENV["RABBITMQ_URL"] || connection_string_or_opts)
+    def initialize(connection_string_or_opts = ENV['RABBITMQ_URL'], optz = Hash.new)
+      opts = case (connection_string_or_opts)
              when nil then
                Hash.new
              when String then
-               self.class.parse_uri(ENV["RABBITMQ_URL"] || connection_string_or_opts)
+               self.class.parse_uri(connection_string_or_opts)
              when Hash then
                connection_string_or_opts
              end.merge(optz)
 
+      @default_hosts_shuffle_strategy = Proc.new { |hosts| hosts.shuffle }
+
       @opts            = opts
-      @host            = self.hostname_from(opts)
-      @port            = self.port_from(opts)
+      log_file         = opts[:log_file] || opts[:logfile] || STDOUT
+      log_level        = opts[:log_level] || ENV["BUNNY_LOG_LEVEL"] || Logger::WARN
+      # we might need to log a warning about ill-formatted IPv6 address but
+      # progname includes hostname, so init like this first
+      @logger          = opts.fetch(:logger, init_default_logger_without_progname(log_file, log_level))
+
+      @addresses       = self.addresses_from(opts)
+      @address_index   = 0
+
+      # re-init, see above
+      @logger          = opts.fetch(:logger, init_default_logger(log_file, log_level))
+
       @user            = self.username_from(opts)
       @pass            = self.password_from(opts)
       @vhost           = self.vhost_from(opts)
-      @logfile         = opts[:log_file] || opts[:logfile] || STDOUT
       @threaded        = opts.fetch(:threaded, true)
 
-      @logger          = opts.fetch(:logger, init_logger(opts[:log_level] || ENV["BUNNY_LOG_LEVEL"] || Logger::WARN))
+      validate_connection_options(opts)
 
       # should automatic recovery from network failures be used?
       @automatically_recover = if opts[:automatically_recover].nil? && opts[:automatic_recovery].nil?
@@ -145,9 +164,11 @@ module Bunny
                                else
                                  opts[:automatically_recover] || opts[:automatic_recovery]
                                end
+      @recovery_attempts     = opts[:recovery_attempts]
       @network_recovery_interval = opts.fetch(:network_recovery_interval, DEFAULT_NETWORK_RECOVERY_INTERVAL)
+      @recover_from_connection_close = opts.fetch(:recover_from_connection_close, true)
       # in ms
-      @continuation_timeout      = opts.fetch(:continuation_timeout, DEFAULT_CONTINUATION_TIMEOUT)
+      @continuation_timeout   = opts.fetch(:continuation_timeout, DEFAULT_CONTINUATION_TIMEOUT)
 
       @status             = :not_connected
       @blocked            = false
@@ -159,7 +180,7 @@ module Bunny
       @channel_max        = @client_channel_max
       @client_heartbeat   = self.heartbeat_from(opts)
 
-      @client_properties   = opts[:properties] || DEFAULT_CLIENT_PROPERTIES
+      @client_properties   = DEFAULT_CLIENT_PROPERTIES.merge(opts.fetch(:properties, {}))
       @mechanism           = opts.fetch(:auth_mechanism, "PLAIN")
       @credentials_encoder = credentials_encoder_for(@mechanism)
       @locale              = @opts.fetch(:locale, DEFAULT_LOCALE)
@@ -171,6 +192,9 @@ module Bunny
       # transport operations/continuations mutex. A workaround for
       # the non-reentrant Ruby mutexes. MK.
       @transport_mutex     = @mutex_impl.new
+      @status_mutex        = @mutex_impl.new
+      @address_index_mutex = @mutex_impl.new
+
       @channels            = Hash.new
 
       @origin_thread       = Thread.current
@@ -179,6 +203,16 @@ module Bunny
       self.initialize_transport
     end
 
+    def validate_connection_options(options)
+      if options[:hosts] && options[:addresses]
+        raise ArgumentError, "Connection options can't contain hosts and addresses at the same time"
+      end
+
+      if (options[:host] || options[:hostname]) && (options[:hosts] || options[:addresses])
+        @logger.warn "The connection options contain both a host and an array of hosts, the single host is ignored."
+      end
+    end
+
     # @return [String] RabbitMQ hostname (or IP address) used
     def hostname;     self.host;  end
     # @return [String] Username used
@@ -208,6 +242,18 @@ module Bunny
       @threaded
     end
 
+    def host
+      @transport ? @transport.host : host_from_address(@addresses[@address_index])
+    end
+
+    def port
+      @transport ? @transport.port : port_from_address(@addresses[@address_index])
+    end
+
+    def reset_address_index
+      @address_index_mutex.synchronize { @address_index = 0 }
+    end
+
     # @private
     attr_reader :mutex_impl
 
@@ -219,52 +265,71 @@ module Bunny
       @transport.configure_socket(&block)
     end
 
+    # @return [Integer] Client socket port
+    def local_port
+      @transport.local_address.ip_port
+    end
+
     # Starts the connection process.
     #
     # @see http://rubybunny.info/articles/getting_started.html
     # @see http://rubybunny.info/articles/connecting.html
     # @api public
     def start
+
       return self if connected?
 
-      @status        = :connecting
+      @status_mutex.synchronize { @status = :connecting }
       # reset here for cases when automatic network recovery kicks in
       # when we were blocked. MK.
       @blocked       = false
       self.reset_continuations
 
       begin
-        # close existing transport if we have one,
-        # to not leak sockets
-        @transport.maybe_initialize_socket
 
-        @transport.post_initialize_socket
-        @transport.connect
+        begin
 
-        if @socket_configurator
-          @transport.configure_socket(&@socket_configurator)
+          # close existing transport if we have one,
+          # to not leak sockets
+          @transport.maybe_initialize_socket
+
+          @transport.post_initialize_socket
+          @transport.connect
+
+          if @socket_configurator
+            @transport.configure_socket(&@socket_configurator)
+          end
+
+          self.init_connection
+          self.open_connection
+
+          @reader_loop = nil
+          self.start_reader_loop if threaded?
+
+        rescue TCPConnectionFailed => e
+          @logger.warn e.message
+          self.initialize_transport
+          @logger.warn "Will try to connect to the next endpoint in line: #{@transport.host}:#{@transport.port}"
+
+          return self.start
+        rescue
+          @status_mutex.synchronize { @status = :not_connected }
+          raise
         end
-
-        self.init_connection
-        self.open_connection
-
-        @reader_loop = nil
-        self.start_reader_loop if threaded?
-
-        @default_channel = self.create_channel
-      rescue Exception => e
-        @status = :not_connected
-        raise e
+      rescue HostListDepleted
+        self.reset_address_index
+        @status_mutex.synchronize { @status = :not_connected }
+        raise TCPConnectionFailedForAllHosts
       end
 
       self
     end
 
-    # Socket operation timeout used by this connection
+    # Socket operation write timeout used by this connection
     # @return [Integer]
     # @private
-    def read_write_timeout
-      @transport.read_write_timeout
+    def transport_write_timeout
+      @transport.write_timeout
     end
 
     # Opens a new channel and returns it. This method will block the calling
@@ -272,14 +337,14 @@ module Bunny
     # opened (this operation is very fast and inexpensive).
     #
     # @return [Bunny::Channel] Newly opened channel
-    def create_channel(n = nil, consumer_pool_size = 1)
+    def create_channel(n = nil, consumer_pool_size = 1, consumer_pool_abort_on_exception = false, consumer_pool_shutdown_timeout = 60)
       raise ArgumentError, "channel number 0 is reserved in the protocol and cannot be used" if 0 == n
 
       @channel_mutex.synchronize do
         if n && (ch = @channels[n])
           ch
         else
-          ch = Bunny::Channel.new(self, n, ConsumerWorkPool.new(consumer_pool_size || 1))
+          ch = Bunny::Channel.new(self, n, ConsumerWorkPool.new(consumer_pool_size || 1, consumer_pool_abort_on_exception, consumer_pool_shutdown_timeout))
           ch.open
           ch
         end
@@ -289,19 +354,18 @@ module Bunny
 
     # Closes the connection. This involves closing all of its channels.
     def close
-      if @transport.open?
-        close_all_channels
+      @status_mutex.synchronize { @status = :closing }
+
+      ignoring_io_errors do
+        if @transport.open?
+          close_all_channels
 
-        Bunny::Timeout.timeout(@transport.disconnect_timeout, ClientTimeout) do
           self.close_connection(true)
         end
 
-        maybe_shutdown_reader_loop
-        close_transport
+        clean_up_on_shutdown
       end
-
-      shut_down_all_consumer_work_pools!
-      @status = :closed
+      @status_mutex.synchronize { @status = :closed }
     end
     alias stop close
 
@@ -325,14 +389,22 @@ module Bunny
       status == :connecting
     end
 
+    # @return [Boolean] true if this AMQP 0.9.1 connection is closing
+    # @api private
+    def closing?
+      @status_mutex.synchronize { @status == :closing }
+    end
+
     # @return [Boolean] true if this AMQP 0.9.1 connection is closed
     def closed?
-      status == :closed
+      @status_mutex.synchronize { @status == :closed }
     end
 
     # @return [Boolean] true if this AMQP 0.9.1 connection is open
     def open?
-      (status == :open || status == :connected || status == :connecting) && @transport.open?
+      @status_mutex.synchronize do
+        (status == :open || status == :connected || status == :connecting) && @transport.open?
+      end
     end
     alias connected? open?
 
@@ -341,40 +413,6 @@ module Bunny
       @automatically_recover
     end
 
-    #
-    # Backwards compatibility
-    #
-
-    # @private
-    def queue(*args)
-      @default_channel.queue(*args)
-    end
-
-    # @private
-    def direct(*args)
-      @default_channel.direct(*args)
-    end
-
-    # @private
-    def fanout(*args)
-      @default_channel.fanout(*args)
-    end
-
-    # @private
-    def topic(*args)
-      @default_channel.topic(*args)
-    end
-
-    # @private
-    def headers(*args)
-      @default_channel.headers(*args)
-    end
-
-    # @private
-    def exchange(*args)
-      @default_channel.exchange(*args)
-    end
-
     # Defines a callback that will be executed when RabbitMQ blocks the connection
     # because it is running low on memory or disk space (as configured via config file
     # and/or rabbitmqctl).
@@ -480,6 +518,7 @@ module Bunny
         raise_if_continuation_resulted_in_a_connection_error!
 
         self.unregister_channel(ch)
+        self.release_channel_id(ch.id)
         @last_channel_close_ok
       end
     end
@@ -503,7 +542,7 @@ module Bunny
 
       shut_down_all_consumer_work_pools!
       maybe_shutdown_heartbeat_sender
-      @status   = :not_connected
+      @status_mutex.synchronize { @status = :not_connected }
     end
 
     # Handles incoming frames and dispatches them.
@@ -515,28 +554,20 @@ module Bunny
     #
     # @private
     def handle_frame(ch_number, method)
-      @logger.debug "Session#handle_frame on #{ch_number}: #{method.inspect}"
+      @logger.debug { "Session#handle_frame on #{ch_number}: #{method.inspect}" }
       case method
       when AMQ::Protocol::Channel::OpenOk then
         @continuations.push(method)
       when AMQ::Protocol::Channel::CloseOk then
         @continuations.push(method)
       when AMQ::Protocol::Connection::Close then
-        @last_connection_error = instantiate_connection_level_exception(method)
-        @continuations.push(method)
-
-        begin
-          shut_down_all_consumer_work_pools!
-          maybe_shutdown_reader_loop
-        rescue ShutdownSignal => sse
-          # no-op
-        rescue Exception => e
-          @logger.warn "Caught an exception when cleaning up after receiving connection.close: #{e.message}"
-        ensure
-          close_transport
+        if recover_from_connection_close?
+          @logger.warn "Recovering from connection.close (#{method.reply_text})"
+          clean_up_on_shutdown
+          handle_network_failure(instantiate_connection_level_exception(method))
+        else
+          clean_up_and_fail_on_connection_close!(method)
         end
-
-        @origin_thread.raise(@last_connection_error)
       when AMQ::Protocol::Connection::CloseOk then
         @last_connection_close_ok = method
         begin
@@ -593,24 +624,34 @@ module Bunny
       end
     end
 
+    # @private
+    def recover_from_connection_close?
+      @recover_from_connection_close
+    end
+
     # @private
     def handle_network_failure(exception)
       raise NetworkErrorWrapper.new(exception) unless @threaded
 
-      @status = :disconnected
+      @status_mutex.synchronize { @status = :disconnected }
 
       if !recovering_from_network_failure?
-        @recovering_from_network_failure = true
-        if recoverable_network_failure?(exception)
-          @logger.warn "Recovering from a network failure..."
-          @channels.each do |n, ch|
-            ch.maybe_kill_consumer_work_pool!
-          end
-          maybe_shutdown_heartbeat_sender
+        begin
+          @recovering_from_network_failure = true
+          if recoverable_network_failure?(exception)
+            @logger.warn "Recovering from a network failure..."
+            @channels.each do |n, ch|
+              ch.maybe_kill_consumer_work_pool!
+            end
+            @reader_loop.stop if @reader_loop
+            maybe_shutdown_heartbeat_sender
 
-          recover_from_network_failure
-        else
-          # TODO: investigate if we can be a bit smarter here. MK.
+            recover_from_network_failure
+          else
+            # TODO: investigate if we can be a bit smarter here. MK.
+          end
+        ensure
+          @recovering_from_network_failure = false
         end
       end
     end
@@ -628,30 +669,39 @@ module Bunny
 
     # @private
     def recover_from_network_failure
-      begin
-        sleep @network_recovery_interval
-        @logger.debug "About to start connection recovery..."
-        self.initialize_transport
-        self.start
+      sleep @network_recovery_interval
+      @logger.debug "About to start connection recovery..."
 
-        if open?
-          @recovering_from_network_failure = false
+      self.initialize_transport
 
-          recover_channels
-        end
-      rescue TCPConnectionFailed, AMQ::Protocol::EmptyResponseError => e
-        @logger.warn "TCP connection failed, reconnecting in #{@network_recovery_interval} seconds"
-        sleep @network_recovery_interval
+      @logger.warn "Retrying connection on next host in line: #{@transport.host}:#{@transport.port}"
+      self.start
+
+      if open?
+        @recovering_from_network_failure = false
+
+        recover_channels
+      end
+    rescue HostListDepleted
+      reset_address_index
+      retry
+    rescue TCPConnectionFailedForAllHosts, TCPConnectionFailed, AMQ::Protocol::EmptyResponseError => e
+      @logger.warn "TCP connection failed, reconnecting in #{@network_recovery_interval} seconds"
+      sleep @network_recovery_interval
+      if should_retry_recovery?
+        @recovery_attempts -= 1 if @recovery_attempts
         retry if recoverable_network_failure?(e)
       end
     end
 
+    # @private
+    def should_retry_recovery?
+      @recovery_attempts.nil? || @recovery_attempts > 1
+    end
+
     # @private
     def recover_channels
-      # default channel is reopened right after connection
-      # negotiation is completed, so make sure we do not try to open
-      # it twice. MK.
-      @channels.reject { |n, ch| ch == @default_channel }.each do |n, ch|
+      @channels.each do |n, ch|
         ch.open
 
         ch.recover_from_network_failure
@@ -687,9 +737,45 @@ module Bunny
       end
     end
 
+    def clean_up_and_fail_on_connection_close!(method)
+      @last_connection_error = instantiate_connection_level_exception(method)
+      @continuations.push(method)
+
+      clean_up_on_shutdown
+      if threaded?
+        @origin_thread.raise(@last_connection_error)
+      else
+        raise @last_connection_error
+      end
+    end
+
+    def clean_up_on_shutdown
+      begin
+        shut_down_all_consumer_work_pools!
+        maybe_shutdown_reader_loop
+        maybe_shutdown_heartbeat_sender
+      rescue ShutdownSignal => sse
+        # no-op
+      rescue Exception => e
+        @logger.warn "Caught an exception when cleaning up after receiving connection.close: #{e.message}"
+      ensure
+        close_transport
+      end
+    end
+
     # @private
-    def hostname_from(options)
-      options[:host] || options[:hostname] || DEFAULT_HOST
+    def addresses_from(options)
+      shuffle_strategy = options.fetch(:hosts_shuffle_strategy, @default_hosts_shuffle_strategy)
+
+      addresses = options[:host] || options[:hostname] || options[:addresses] ||
+        options[:hosts] || ["#{DEFAULT_HOST}:#{port_from(options)}"]
+      addresses = [addresses] unless addresses.is_a? Array
+
+      addresses.map! do |address|
+        host_with_port?(address) ? address : "#{address}:#{port_from(@opts)}"
+      end
+
+      shuffle_strategy.call addresses
     end
 
     # @private
@@ -703,6 +789,63 @@ module Bunny
       options.fetch(:port, fallback)
     end
 
+    # @private
+    def host_with_port?(address)
+      # we need to handle cases such as [2001:db8:85a3:8d3:1319:8a2e:370:7348]:5671
+      last_colon                  = address.rindex(":")
+      last_closing_square_bracket = address.rindex("]")
+
+      if last_closing_square_bracket.nil?
+        address.include?(":")
+      else
+        last_closing_square_bracket < last_colon
+      end
+    end
+
+    # @private
+    def host_from_address(address)
+      # we need to handle cases such as [2001:db8:85a3:8d3:1319:8a2e:370:7348]:5671
+      last_colon                  = address.rindex(":")
+      last_closing_square_bracket = address.rindex("]")
+
+      if last_closing_square_bracket.nil?
+        parts = address.split(":")
+        # this looks like an unquoted IPv6 address, so emit a warning
+        if parts.size > 2
+          @logger.warn "Address #{address} looks like an unquoted IPv6 address. Make sure you quote IPv6 addresses like so: [2001:db8:85a3:8d3:1319:8a2e:370:7348]"
+        end
+        return parts[0]
+      end
+
+      if last_closing_square_bracket < last_colon
+        # there is a port
+        address[0, last_colon]
+      elsif last_closing_square_bracket > last_colon
+        address
+      end
+    end
+
+    # @private
+    def port_from_address(address)
+      # we need to handle cases such as [2001:db8:85a3:8d3:1319:8a2e:370:7348]:5671
+      last_colon                  = address.rindex(":")
+      last_closing_square_bracket = address.rindex("]")
+
+      if last_closing_square_bracket.nil?
+        parts = address.split(":")
+        # this looks like an unquoted IPv6 address, so emit a warning
+        if parts.size > 2
+          @logger.warn "Address #{address} looks like an unquoted IPv6 address. Make sure you quote IPv6 addresses like so: [2001:db8:85a3:8d3:1319:8a2e:370:7348]"
+        end
+        return parts[1].to_i
+      end
+
+      if last_closing_square_bracket < last_colon
+        # there is a port
+        address[(last_colon + 1)..-1].to_i
+      end
+    end
+
     # @private
     def vhost_from(options)
       options[:virtual_host] || options[:vhost] || DEFAULT_VHOST
@@ -767,7 +910,7 @@ module Bunny
         if threaded?
           # this is the easiest way to wait until the loop
           # is guaranteed to have terminated
-          @reader_loop.raise(ShutdownSignal)
+          @reader_loop.terminate_with(ShutdownSignal)
           # joining the thread here may take forever
           # on JRuby because sun.nio.ch.KQueueArrayWrapper#kevent0 is
           # a native method that cannot be (easily) interrupted.
@@ -836,7 +979,7 @@ module Bunny
       end
     end
 
-    # Sends multiple frames, one by one. For thread safety this method takes a channel
+    # Sends multiple frames, in one go. For thread safety this method takes a channel
     # object and synchronizes on it.
     #
     # @private
@@ -847,7 +990,9 @@ module Bunny
       # If we synchronize on the channel, however, this is both thread safe and pretty fine-grained
       # locking. Note that "single frame" methods do not need this kind of synchronization. MK.
       channel.synchronize do
-        frames.each { |frame| self.send_frame(frame, false) }
+        # see rabbitmq/rabbitmq-server#156
+        data = frames.reduce("") { |acc, frame| acc << frame.encode }
+        @transport.write(data)
         signal_activity!
       end
     end # send_frameset(frames)
@@ -885,7 +1030,12 @@ module Bunny
     # @return [String]
     # @api public
     def to_s
-      "#<#{self.class.name}:#{object_id} #{@user}@#{@host}:#{@port}, vhost=#{@vhost}>"
+      oid = ("0x%x" % (self.object_id << 1))
+      "#<#{self.class.name}:#{oid} #{@user}@#{host}:#{port}, vhost=#{@vhost}, addresses=[#{@addresses.join(',')}]>"
+    end
+
+    def inspect
+      to_s
     end
 
     protected
@@ -902,7 +1052,7 @@ module Bunny
       @server_authentication_mechanisms = (connection_start.mechanisms || "").split(" ")
       @server_locales                   = Array(connection_start.locales)
 
-      @status = :connected
+      @status_mutex.synchronize { @status = :connected }
     end
 
     # @private
@@ -918,13 +1068,11 @@ module Bunny
                 fr
                 # frame timeout means the broker has closed the TCP connection, which it
                 # does per 0.9.1 spec.
-              rescue Errno::ECONNRESET, ClientTimeout, AMQ::Protocol::EmptyResponseError, EOFError, IOError => e
+              rescue
                 nil
               end
       if frame.nil?
-        @state = :closed
-        @logger.error "RabbitMQ closed TCP connection before AMQP 0.9.1 connection was finalized. Most likely this means authentication failure."
-        raise Bunny::PossibleAuthenticationFailureError.new(self.user, self.vhost, self.password.size)
+        raise TCPConnectionFailed.new('An empty frame was received while opening the connection. In RabbitMQ <= 3.1 this could mean an authentication issue.')
       end
 
       response = frame.decode_payload
@@ -946,15 +1094,29 @@ module Bunny
                               else
                                 negotiate_value(@client_heartbeat, connection_tune.heartbeat)
                               end
-      @logger.debug "Heartbeat interval negotiation: client = #{@client_heartbeat}, server = #{connection_tune.heartbeat}, result = #{@heartbeat}"
+      @logger.debug { "Heartbeat interval negotiation: client = #{@client_heartbeat}, server = #{connection_tune.heartbeat}, result = #{@heartbeat}" }
       @logger.info "Heartbeat interval used (in seconds): #{@heartbeat}"
 
-      @channel_id_allocator = ChannelIdAllocator.new(@channel_max)
+      # We set the read_write_timeout to twice the heartbeat value
+      # This allows us to miss a single heartbeat before we time out the socket.
+      @transport.read_timeout = if heartbeat_disabled?(@client_heartbeat)
+                                  Transport::DEFAULT_READ_TIMEOUT
+                                else
+                                  # pad to account for edge cases. MK.
+                                  @heartbeat * 2.2
+                                end
+
+
+      # if there are existing channels we've just recovered from
+      # a network failure and need to fix the allocated set. See issue 205. MK.
+      if @channels.empty?
+        @channel_id_allocator = ChannelIdAllocator.new(@channel_max)
+      end
 
       @transport.send_frame(AMQ::Protocol::Connection::TuneOk.encode(@channel_max, @frame_max, @heartbeat))
-      @logger.debug "Sent connection.tune-ok with heartbeat interval = #{@heartbeat}, frame_max = #{@frame_max}, channel_max = #{@channel_max}"
+      @logger.debug { "Sent connection.tune-ok with heartbeat interval = #{@heartbeat}, frame_max = #{@frame_max}, channel_max = #{@channel_max}" }
       @transport.send_frame(AMQ::Protocol::Connection::Open.encode(self.vhost))
-      @logger.debug "Sent connection.open with vhost = #{self.vhost}"
+      @logger.debug { "Sent connection.open with vhost = #{self.vhost}" }
 
       frame2 = begin
                  fr = @transport.read_next_frame
@@ -964,17 +1126,15 @@ module Bunny
                  fr
                  # frame timeout means the broker has closed the TCP connection, which it
                  # does per 0.9.1 spec.
-               rescue Errno::ECONNRESET, ClientTimeout, AMQ::Protocol::EmptyResponseError, EOFError => e
+               rescue
                  nil
                end
       if frame2.nil?
-        @state = :closed
-        @logger.warn "RabbitMQ closed TCP connection before AMQP 0.9.1 connection was finalized. Most likely this means authentication failure."
-        raise Bunny::PossibleAuthenticationFailureError.new(self.user, self.vhost, self.password.size)
+        raise TCPConnectionFailed.new('An empty frame was received while opening the connection. In RabbitMQ <= 3.1 this could mean an authentication issue.')
       end
       connection_open_ok = frame2.decode_payload
 
-      @status = :open
+      @status_mutex.synchronize { @status = :open }
       if @heartbeat && @heartbeat > 0
         initialize_heartbeat_sender
       end
@@ -993,7 +1153,11 @@ module Bunny
             close_transport
           end
 
-          @origin_thread.raise(e)
+          if threaded?
+            @origin_thread.raise(e)
+          else
+            raise e
+          end
         else
           raise "could not open connection: server did not respond with connection.open-ok but #{connection_open_ok.inspect} instead"
         end
@@ -1030,7 +1194,22 @@ module Bunny
 
     # @private
     def initialize_transport
-      @transport = Transport.new(self, @host, @port, @opts.merge(:session_thread => @origin_thread))
+      if address = @addresses[ @address_index ]
+        @address_index_mutex.synchronize { @address_index += 1 }
+        @transport.close rescue nil # Let's make sure the previous transport socket is closed
+        @transport = Transport.new(self,
+                                   host_from_address(address),
+                                   port_from_address(address),
+                                   @opts.merge(:session_thread => @origin_thread)
+                                  )
+
+        # Reset the cached progname for the logger only when no logger was provided
+        @default_logger.progname = self.to_s
+
+        @transport
+      else
+        raise HostListDepleted
+      end
     end
 
     # @private
@@ -1078,12 +1257,22 @@ module Bunny
     end
 
     # @private
-    def init_logger(level)
-      lgr          = ::Logger.new(@logfile)
-      lgr.level    = normalize_log_level(level)
-      lgr.progname = self.to_s
+    def init_default_logger(logfile, level)
+      @default_logger = begin
+                          lgr = ::Logger.new(logfile)
+                          lgr.level    = normalize_log_level(level)
+                          lgr.progname = self.to_s
+                          lgr
+                        end
+    end
 
-      lgr
+    # @private
+    def init_default_logger_without_progname(logfile, level)
+      @default_logger = begin
+                          lgr = ::Logger.new(logfile)
+                          lgr.level    = normalize_log_level(level)
+                          lgr
+                        end
     end
 
     # @private
@@ -1116,6 +1305,14 @@ module Bunny
         n
       end
     end
+
+    def ignoring_io_errors(&block)
+      begin
+        block.call
+      rescue AMQ::Protocol::EmptyResponseError, IOError, SystemCallError, Bunny::NetworkFailure => _
+        # ignore
+      end
+    end
   end # Session
 
   # backwards compatibility
