@@ -44,11 +44,11 @@ describe Bunny::Queue, "#subscribe_with" do
       end
       t.abort_on_exception = true
 
-      q1.subscribe_with(ec, :ack => true)
+      q1.subscribe_with(ec, :manual_ack => true)
       sleep 2
       ch1.close
 
-      q2.message_count.should == 50
+      expect(q2.message_count).to eq 50
     end
   end
 end
