@@ -1,23 +1,30 @@
 # -*- coding: utf-8 -*-
 require "spec_helper"
 
-unless ENV["CI"]
-  describe "x-consistent-hash exchange" do
-    let(:connection) do
-      c = Bunny.new(:user => "bunny_gem", :password => "bunny_password", :vhost => "bunny_testbed")
-      c.start
-      c
-    end
+describe "x-consistent-hash exchange" do
+  let(:connection) do
+    c = Bunny.new(:user => "bunny_gem", :password => "bunny_password", :vhost => "bunny_testbed",
+      :automatically_recover => true)
+    c.start
+    c
+  end
 
-    after :each do
-      connection.close
-    end
+  let(:http_client) { RabbitMQ::HTTP::Client.new("http://127.0.0.1:15672") }
 
-    let(:list) { Range.new(0, 6).to_a.map(&:to_s) }
+  after :each do
+    connection.close
+  end
 
-    let(:m) { 1500 }
+  let(:list) { Range.new(0, 6).to_a.map(&:to_s) }
 
-    it "distributes messages between queues bound with the same routing key" do
+  let(:m) { 1500 }
+
+  let(:exchange_plugin_available) do
+    http_client.overview.exchange_types.map(&:name).include?("x-consistent-hash")
+  end
+
+  it "distributes messages between queues bound with the same routing key" do
+    if exchange_plugin_available
       ch   = connection.create_channel
       body = "сообщение"
       # requires the consistent hash exchange plugin,
@@ -41,10 +48,12 @@ unless ENV["CI"]
       end
 
       sleep 2.0
-      q1.message_count.should be > 100
-      q2.message_count.should be > 100
+      expect(q1.message_count).to be > 100
+      expect(q2.message_count).to be > 100
 
       ch.close
-    end
-  end
+    else
+      skip "x-consistent-hash exchange type isn't available"
+    end # if
+  end # it
 end
