@@ -22,13 +22,17 @@ describe "A message" do
     x.publish("")
     sleep 0.2
 
-    delivery_info, _, _ = q.pop(:ack => true)
-    dlq.message_count.should be_zero
+    delivery_info, _, _ = q.pop(:manual_ack => true)
+    expect(dlq.message_count).to be_zero
     ch.nack(delivery_info.delivery_tag)
 
     sleep 0.2
-    q.message_count.should be_zero
-    dlq.message_count.should == 1
+    expect(q.message_count).to be_zero
+
+    delivery, properties, body = dlq.pop
+    ds = properties.headers["x-death"]
+    expect(ds).not_to be_empty
+    expect(ds.first["reason"]).to eq("rejected")
 
     dlx.delete
   end
@@ -44,8 +48,27 @@ describe "A message" do
     x.publish("")
     sleep 0.2
 
-    q.message_count.should be_zero
-    dlq.message_count.should == 1
+    expect(q.message_count).to be_zero
+    expect(dlq.message_count).to eq 1
+
+    dlx.delete
+  end
+
+  it "carries the x-death header" do
+    ch   = connection.create_channel
+    x    = ch.fanout("amq.fanout")
+    dlx  = ch.fanout("bunny.tests.dlx.exchange")
+    q    = ch.queue("", :exclusive => true, :arguments => {"x-dead-letter-exchange" => dlx.name, "x-message-ttl" => 100}).bind(x)
+    # dead letter queue
+    dlq  = ch.queue("", :exclusive => true).bind(dlx)
+
+    x.publish("")
+    sleep 0.2
+
+    delivery, properties, body = dlq.pop
+    ds = properties.headers["x-death"]
+    expect(ds).not_to be_empty
+    expect(ds.first["reason"]).to eq("expired")
 
     dlx.delete
   end
