@@ -24,14 +24,14 @@ describe Bunny::Exchange do
     source.publish("")
     sleep 0.5
 
-    queue.message_count.should be == 1
-    queue.pop(:ack => true)
+    expect(queue.message_count).to eq 1
+    queue.pop(:manual_ack => true)
 
     destination.unbind(source)
     source.publish("")
     sleep 0.5
 
-    queue.message_count.should be == 0
+    expect(queue.message_count).to eq 0
 
     source.delete
     destination.delete
