@@ -19,12 +19,12 @@ describe Bunny::Queue, "bound to an exchange" do
 
     x.publish("")
     sleep 0.3
-    q.message_count.should == 1
+    expect(q.message_count).to eq 1
 
     q.unbind(x)
 
     x.publish("")
-    q.message_count.should == 1
+    expect(q.message_count).to eq 1
   end
 end
 
@@ -42,7 +42,7 @@ describe Bunny::Queue, "NOT bound to an exchange" do
   end
 
 
-  it "cannot be unbound (raises a channel error)" do
+  it "is idempotent (succeeds)" do
     ch = connection.create_channel
     x  = ch.fanout("amq.fanout")
     q  = ch.queue("", :exclusive => true)
