@@ -1,19 +1,18 @@
 require "spec_helper"
 
 describe Bunny::Channel, "#open" do
-  let(:connection) do
-    c = Bunny.new(:user => "bunny_gem", :password => "bunny_password", :vhost => "bunny_testbed")
-    c.start
-    c
+  before :all do
+    @connection = Bunny.new(:user => "bunny_gem", :password => "bunny_password", :vhost => "bunny_testbed")
+    @connection.start
   end
 
   after :all do
-    connection.close if connection.open?
+    @connection.close if @connection.open?
   end
 
 
   it "properly resets channel exception state" do
-    ch = connection.create_channel
+    ch = @connection.create_channel
 
     begin
       ch.queue("bunny.tests.does.not.exist", :passive => true)
