@@ -12,27 +12,39 @@ unless defined?(JRUBY_VERSION) && !ENV["FORCE_JRUBY_RUN"]
         end
 
     n.times do |i|
-      it "can be closed (take #{i})" do
+      it "can be closed (automatic recovery disabled, take #{i})" do
         c  = Bunny.new(:automatically_recover => false)
         c.start
         ch = c.create_channel
 
-        c.should be_connected
+        expect(c).to be_connected
         c.stop
-        c.should be_closed
+        expect(c).to be_closed
+      end
+    end
+
+    n.times do |i|
+      it "can be closed (automatic recovery enabled, take #{i})" do
+        c  = Bunny.new(:automatically_recover => true)
+        c.start
+        ch = c.create_channel
+
+        expect(c).to be_connected
+        c.stop
+        expect(c).to be_closed
       end
     end
 
     context "in the single threaded mode" do
       n.times do |i|
-        it "can be closed (take #{i})" do
+        it "can be closed (single threaded mode, take #{i})" do
           c  = Bunny.new(:automatically_recover => false, :threaded => false)
           c.start
           ch = c.create_channel
 
-          c.should be_connected
+          expect(c).to be_connected
           c.stop
-          c.should be_closed
+          expect(c).to be_closed
         end
       end
     end
